// Deklarationsdatei einbinden
#include "tspIndividuum.h"

// Konstruktor
tspIndividuum::tspIndividuum()
    : cppvmObject()
{
  xPositionen = new float[TSP_ANZAHL_LAEDEN];
  yPositionen = new float[TSP_ANZAHL_LAEDEN];
  mutationStreckenLaenge=TSP_MUTATION_STRECKEN_LAENGE;
  routenLaenge=0.0;
}

// Destruktor
tspIndividuum::~tspIndividuum()
{
  delete [] xPositionen;
  delete [] yPositionen;
}

// In dieser Methode muessen alle Daten angegeben werden,
// die beim Datentransfer beruecksichtigt werden sollen.
void
tspIndividuum::cppvmTransfer()
{
  // Transfer von C++ Standard Typen (int, float, double, ...)
  // mit den Funktionen 'cppvmTransfer<Type>'
  cppvmTransferInt(&mutationStreckenLaenge);
  cppvmTransferFloat(&routenLaenge);
  cppvmTransferFloat(xPositionen, TSP_ANZAHL_LAEDEN);
  cppvmTransferFloat(yPositionen, TSP_ANZAHL_LAEDEN);
}

tspIndividuum&
tspIndividuum::operator=(const tspIndividuum& obj)
{
  mutationStreckenLaenge = obj.mutationStreckenLaenge;
  routenLaenge = obj.routenLaenge;
  int i;
  for(i=0; i<TSP_ANZAHL_LAEDEN; i++){
    xPositionen[i] = obj.xPositionen[i];
    yPositionen[i] = obj.yPositionen[i];
  }
  return *this;
}

float
tspIndividuum::zufallswert(float wertebereichMin, float wertebereichMax)
{
  // Sicherheitsabfrage
  if(wertebereichMin<0 || wertebereichMax<0 ||
     wertebereichMin>wertebereichMax){
    cerr << "zufallswert: Wertebereiche sind falsch gesetzt" << endl;
    return 0.0;
  }

  static bool ersterAufruf=true;
  if(ersterAufruf){
    ersterAufruf=false;
    // Zufallsgenerator mit Zeitabfrage initialisieren
    srand(ctSeed());
  }
  
  // Zufallswert im vorgegebenen Wertebereich erzeugen 
  return (float)(wertebereichMin + 
          ((wertebereichMax*(float)rand())/((float)RAND_MAX+1.0)));
}

void
tspIndividuum::init(float wertebereichXMin, float wertebereichXMax,
                    float wertebereichYMin, float wertebereichYMax)
{
  int i;
  for(i=0; i<TSP_ANZAHL_LAEDEN; i++){
    xPositionen[i] = (float)
      ((int)zufallswert(wertebereichXMin, wertebereichXMax));
    yPositionen[i] = (float)
      ((int)zufallswert(wertebereichYMin, wertebereichYMax));
  }
  routeBerechnen();
}

bool
tspIndividuum::liesRouteVonDatei(char *dateiname)
{
  if(dateiname==NULL){
    cerr << "liesRouteVonDatei; dateiname is NULL." << endl;
    return false;
  }
    
  FILE *fp;
  fp = fopen(dateiname, "r");
  if(fp==NULL){
    cerr << "liesRouteVonDatei; Konnte Datei "
         << dateiname << " nicht oeffnen." << endl;
    return false;
  }

  int i=0;
  bool leseFehler=false;
  int geleseneWerte;

  while(i<TSP_ANZAHL_LAEDEN && leseFehler==false){
    geleseneWerte = fscanf(fp, "%f %f\n", &xPositionen[i], &yPositionen[i]);
    i++;
    if(geleseneWerte!=2){
      leseFehler = true;
    }
  }
  
  fclose(fp);

  // neue Routenlaenge berechnen
  routeBerechnen();

  if(leseFehler){
    return false;
  }
  else{
    return true;
  }
}

bool
tspIndividuum::schreibeRouteInDatei(char *dateiname)
{
  if(dateiname==NULL){
    cerr << "schreibeRouteInDatei; dateiname is NULL." << endl;
    return false;
  }
    
  FILE *fp;
  fp = fopen(dateiname, "w");
  if(fp==NULL){
    cerr << "schreibeRouteInDatei; Konnte Datei "
         << dateiname << " nicht oeffnen." << endl;
    return false;
  }

  int i;

  // alle Positionspaare schreiben
  for(i=0; i<TSP_ANZAHL_LAEDEN;i++){
    fprintf(fp, "%f %f\n", xPositionen[i], yPositionen[i]);
  }
  // zusaetzlich erste Position schreiben, um
  // bei Gnuplot die letzte Position mit der ersten
  // verbinden zu koennen
  fprintf(fp, "%f %f\n", xPositionen[0], yPositionen[0]);
  
  fclose(fp);
  return true;
}

void
tspIndividuum::mutation()
{
  //////////////////////////////////
  // Mutation der Streckenlaenge //
  ////////////////////////////////

  // mutieren?
  if(zufallswert(0.0, 1.0)<=TSP_MUTATIONSRATE){
    // Mutieren plus oder minus 
    if((int)zufallswert(0.0, 2.0)==0){
      mutationStreckenLaenge = mutationStreckenLaenge -
        TSP_MUTATIONSSCHRITTWEITE;
      if(mutationStreckenLaenge<2){
        // das Vertauschen der Strecke ist erst ab 2 Positionen sinnvoll
        mutationStreckenLaenge=2;
      }
    }
    else{
      mutationStreckenLaenge = mutationStreckenLaenge +
        TSP_MUTATIONSSCHRITTWEITE;
      if(mutationStreckenLaenge>TSP_ANZAHL_LAEDEN-2){
        // das Vertauschen der Strecke ist maximal bis
        // zu TSP_ANZAHL_LAEDEN-2 Positionen sinnvoll
        mutationStreckenLaenge=TSP_ANZAHL_LAEDEN-2;
      }
    }
  }

  ////////////////////////////////////////
  // Mutation eines Streckenabschnitts //
  //////////////////////////////////////

  // Zufallszahl zwichen 0 und 'TSP_ANZAHL_LAEDEN'
  int schnittposition = (int)zufallswert(0.0, (float)TSP_ANZAHL_LAEDEN);
  
  // ab 'schnittpostion' bis 'schnittposition+mutationStreckenLaenge'
  // invertieren der Positionen; dabei Ueberschreitungen der Routen-
  // Raender beachten
  int posLinks = schnittposition;
  int posRechts = (schnittposition+mutationStreckenLaenge-1) % 
    TSP_ANZAHL_LAEDEN;
  float tmpX, tmpY;
  int i;

  for(i=0; i<mutationStreckenLaenge/2; i++){
    // Vertauschen der Ladenpositionen
    tmpX = xPositionen[posLinks];
    tmpY = yPositionen[posLinks];
    xPositionen[posLinks]=xPositionen[posRechts];
    yPositionen[posLinks]=yPositionen[posRechts];
    xPositionen[posRechts]=tmpX;
    yPositionen[posRechts]=tmpY;
    // neue Vertauschkandidaten bestimmen
    // Raender beachten!
    posLinks++;
    if(posLinks==TSP_ANZAHL_LAEDEN){
      posLinks=0;
    }
    posRechts--;
    if(posRechts==-1){
      posRechts=TSP_ANZAHL_LAEDEN-1;
    }
  }
  // neue Routenlaenge berechnen
  routeBerechnen();
}

float
tspIndividuum::distanz(unsigned int index1, unsigned int index2)
{
  // Sicherheitsabfrage
  if(index1>TSP_ANZAHL_LAEDEN-1 ||
     index2>TSP_ANZAHL_LAEDEN-1){
    cerr << "distanz: Indizes sind falsch gesetzt" << endl;
    return 0.0;
  }

  double a, b;
  a = fabs(xPositionen[index1]-xPositionen[index2]);
  b = fabs(yPositionen[index1]-yPositionen[index2]);
  
  return (float)(sqrt((a*a)+(b*b)));
}

void
tspIndividuum::routeBerechnen()
{
  routenLaenge=0.0;
  int i;
  for(i=0; i<TSP_ANZAHL_LAEDEN-1; i++){
    routenLaenge = routenLaenge + distanz(i, i+1);
  }
  // Route vom letzten zum ersten Laden nicht vergessen
  routenLaenge = routenLaenge + distanz(0,TSP_ANZAHL_LAEDEN-1);
}

// Vergleichsfunktion fuer cppvmObjectCollection Liste
bool tspIndividuumVergleich(cppvmObject* obj1, cppvmObject* obj2)
{
  if(((tspIndividuum*)obj1)->gibRoutenLaenge() <
     ((tspIndividuum*)obj2)->gibRoutenLaenge()){
    return true;
  }
  else{
    return false;
  }
}

  
