// CPPvm bekannt machen
#include "cppvm.h"
// Route
#include "tspIndividuum.h"
// Defines/Makros
#include "ctdefines.h"
#include "tspdefines.h"

bool gnuplotAn = false;

void
ergebnisAusgabe(tspIndividuum& besteRoute, 
                int taskId, bool beenden=false)
{
  static FILE *pipeErgGnuplot;
  static FILE *pipeBesteRouteGnuplot;
  static FILE *dateiErg;
  static double startTime;
  static int aufrufe;

  static bool ersterAufruf=true;
  if(ersterAufruf){
    startTime = CT_SYSTEM_TIME;
    aufrufe=0;
    if(gnuplotAn){
      pipeErgGnuplot = CT_POPEN(CT_GNUPLOT_EXE, "w");
      pipeBesteRouteGnuplot = CT_POPEN(CT_GNUPLOT_EXE, "w");
      if(pipeErgGnuplot==NULL || pipeBesteRouteGnuplot==NULL){
        cerr << "Visualisierung: konnte gnuplot nicht oeffnen!" << endl;
        return;
      }
      fprintf(pipeErgGnuplot, "set xlabel \"Zeit(s)\" 0,-1\n");
      fprintf(pipeErgGnuplot, "set ylabel \"Routen-Laenge\" -1,0\n");
      fflush(pipeErgGnuplot);
      fprintf(pipeBesteRouteGnuplot, "set xlabel \"Beste Route\" 0,-1\n");
      fflush(pipeBesteRouteGnuplot);
    }
    dateiErg = fopen("tspErgebnisse.txt","w");
    if(dateiErg==NULL){
      cerr << "Visualisierung: konnte Datei tspErgebnisse.txt nicht oeffnen!" 
           << endl;
      return;
    }
    ersterAufruf=false;
  }
  if(beenden==true){
    ersterAufruf=true;
    fclose(dateiErg);
    if(gnuplotAn){
      CT_PCLOSE(pipeErgGnuplot);
      CT_PCLOSE(pipeBesteRouteGnuplot);
    }
    return;
  }

  // Beste Route in Dateien speichern
  besteRoute.schreibeRouteInDatei("tspBesteRoute.txt");
  fprintf(dateiErg, "%f %f\n", CT_SYSTEM_TIME-startTime, 
          besteRoute.gibRoutenLaenge());
  fflush(dateiErg);
  
  if(gnuplotAn){
    // Route und Laengen in Gnuplot ausgeben
    fprintf(pipeErgGnuplot, 
            "plot \"tspErgebnisse.txt\" with linespoints \n");
    fflush(pipeErgGnuplot);
    fprintf(pipeBesteRouteGnuplot, 
            "plot \"tspBesteRoute.txt\" with linespoints \n");
    fflush(pipeBesteRouteGnuplot);
  }

  aufrufe++;
  cout << "Num: " << aufrufe
       << " Zeit(s): " 
       << CT_SYSTEM_TIME-startTime
       << " Laenge: " 
       << besteRoute.gibRoutenLaenge()
       << " von Prozess: " 
       << taskId
       << " Mutations-Streckenlaenge: " 
       << besteRoute.gibMutationStreckenLaenge()
       << endl;
  
}

int
main(int argc, char *argv[])
{

  // Werden dem Programm Argumente uebergeben, so wird gnuplot
  // angeschaltet; die Werte der Argumente sind ohne Bedeutung
  if(argc>1){
    gnuplotAn=true;
  }

  // Individuen (Routen)
  tspIndividuum besteRoute;
  tspIndividuum neueRoute;
  int taskId=0;

  // Vorgegebene Route von Datei lesen, sonst Zufallswerte verwenden
  if(besteRoute.liesRouteVonDatei(TSP_ANFANGS_ROUTE_DATEI)==false){
    besteRoute.init();
  }
  
  // Den eigenen Prozess in die "Parallel 
  // Virtual Machine" (PVM) einklinken
  cppvmConnection pvmConn;

  // Hinzunahmen zur Verfuegung stehender Rechenknoten
  //pvmConn.addHost("sun1");
  //pvmConn.addHost("linux1");
  //pvmConn.addHost("win1");

  // Starten der Prozesse "tspRechenknecht".
  // Die Auswahl der Rechner wird PVM ueberlassen.
  cppvmSpawnMultiConnection rechenknechte("tspRechenknecht", 
                                          TSP_ANZAHL_RECHENKNECHTE);

  // Beteiligte Rechner ausgeben
  pvmConn.printHostInfo();
  // Beteiligte Prozesse ausgeben
  pvmConn.printTaskInfo();
  
  // Nachrichtenkanal:
  // nicht blockierendes Senden an all Rechenknechte
  cppvmSendStream sendStrm(rechenknechte, CPPvmSNBallChildren);
  // Initiale Route an alle Rechenknechte
  sendStrm << besteRoute;

  // Initiale Route ausgeben
  ergebnisAusgabe(besteRoute, taskId);

  // Blockierendes Empfangen von irgendeinem Prozess
  cppvmReceiveStream recStrm(CPPvmRBany);

  int i;
  for(i=0; i<TSP_ANZAHL_DURCHLAUFE; i++){
    recStrm >> neueRoute;
    if(neueRoute.gibRoutenLaenge()<besteRoute.gibRoutenLaenge()){
      besteRoute=neueRoute;
      taskId = recStrm.getLastRecFromTid();
    }
    // Senden der besten Route an den Prozess, der gerade
    // die neue Route geliefert hat
    sendStrm.set(recStrm, CPPvmSNBback);
    sendStrm << besteRoute;

    ergebnisAusgabe(besteRoute, taskId);
  }

  cout << "Return zum Beenden druecken..." << endl;
  getchar();

  // Dateien schliessen
  ergebnisAusgabe(besteRoute, taskId, true);

  // Das wars, beenden aller beteiligten Prozesse
  pvmConn.halt();

  return 0;
}
